//
// Epxrlcorr.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Class to correct an Epic-pn ODF for the effects of 
//		X-ray loading. Translated from an original IDL script
//		written by Matteo Guainazzi.
//
// Method:
//
//           Loop over each *_*_PN*IME.FIT (raw events per CCD) file
//
//                Identify and read the offset map
//
//                Calculate XRL by subtracting the median of the columns
//                and then the median of the rows
//
//                Create a XRL map
//
//                Apply the correction to the energy of each raw charge
//
//                close file
//
//           End of loop
//
#include <iostream>
#include "OalServer.h"
#include "CalServer.h"
#include "Dal.h"

#include "Epxrlcorr.h"

//#include "CalSetter.h"

#include "Tracer.h"
#include "errstr.h"
#include "param.h"
#include "slatec.h"
#include "Image.h"
//#include "OdfObject.h"

/* Method: Epxrlcorr() - Constructor */
Epxrlcorr::Epxrlcorr()
{
  PROC;

  /* Check that the raw PN event file exists within the ODF and is ok */
  readRawEventFile(stringParameter("eventset"));

  /* Open the input file for writing or clone the input file into the 
     output file */
  DataSet *evset = 0;

  if (booleanParameter("withoutset"))
     evset = dataSetServer->clone(stringParameter("eventset"),
                                  stringParameter("outset"), Dal::Modify);
  else {
     evset = dataSetServer->open(stringParameter("eventset"), DataSet::Modify);
  }
 
  /* Find the name of the offset map */
  string offsetmap = findOffsetMap();

  /* Use the offset map to calculate the X-ray loading factors */
  SImage XRLCorrMap = calculateXRL(offsetmap);

  /* Apply the XRL correction to the charges in the raw event file */
  XRLCorrApply(evset, XRLCorrMap);

  /* Close down and go home */
  dataSetServer->close(evset);

  /* Output success message */
  errstr<<"X-ray loading corrections completed "
        <<Sas::message(Msg::App, Msg::Sparse);

}

Epxrlcorr::~Epxrlcorr()
{
}

/* Method: readRawEventFile()
*  Purpose: opens the raw event file and checks contents are ok
*/
void Epxrlcorr::readRawEventFile(string eventset) 
{

  DataSet * evset=0;

 // Check that the dataset exists
  if (dataSetServer->exists(eventset)) {
      evset = dataSetServer->open(eventset, DataSet::Modify);

    // Check that the event extension exists
      Table * evtab=0;
      if (evset->hasBlock("PNIME1")) { // Raw event file
         evtab = evset->table("PNIME1");
         oalServer.setStateFromSetName(eventset);
         _rawinputfile=true;
      }
      else if (evset->hasBlock("PNTIE1") || evset->hasBlock("PNBUE1"))
         errstr<<" Can NOT perform X-ray loading correction on "
               <<" Timing or Burst mode file "<<eventset
               <<" with this task. Please use epreject."
               << Sas::error("UnsuportedObservingMode"); 

      else if (evset->hasBlock("EVENTS")) { // Output of epframes

         errstr<<"EPFRAMES input file"
               <<Sas::message(Msg::App, Msg::Verbose);

         evtab = evset->table("EVENTS");
                  
        // Read attributes from input file and set the OAL server
         if (evtab->hasAttribute("QUADRANT") && evtab->hasAttribute("CCDID")
             && evtab->hasAttribute("EXPIDSTR")) {

            int ccdnr = evtab->attribute("QUADRANT")->asInt() * 3 +
                 evtab->attribute("CCDID")->asInt() + 1;

            oalServer.instrument(Xmm::EPN);
            oalServer.ccd(ccdnr);
            oalServer.dataMode(Xmm::IMAGING);
            oalServer.exposureId(evtab->attribute("EXPIDSTR")->asString());

           // Set flag to show that input file is from epframes
            _rawinputfile=false;
         }

         else
            errstr<<"QUADRANT or CCDID or EXPIDSTR keyword not found in "
                  <<"EVENTS extension of dataset "<<eventset
                  << Sas::error("MissingKeyword");
      }
      else
         errstr<<" PNIME1 block not found in dataset "<<eventset
               << Sas::error("NoEventBlock"); 

    // Check whether this eventset has already been corrected
      if (evtab->hasAttribute("XRL_CORR"))
         if (evtab->attribute("XRL_CORR")->asString() == "YES")
            errstr<<" Dataset: "<< eventset
                  <<" has already been corrected for X-ray loading"
                  << Sas::error("XRLCorrected");

      if (evtab->hasAttribute("REJ_CORR"))
         if (evtab->attribute("REJ_CORR")->asString() == "YES")
            errstr<<" Dataset: "<< eventset
                  <<" has already been corrected for X-ray loading"
                  << Sas::error("XRLCorrected");

    // If everything ok find the CCD number for this event set 
/*
      if (evtab->hasAttribute("QUADRANT") && evtab->hasAttribute("CCDID"))
//         _ccdnr = evtab->attribute("QUADRANT")->asInt() * 3 +
  //                evtab->attribute("CCDID")->asInt() + 1;
      else
         errstr<<"QUADRANT or CCDID keyword not found in EVENTS extension "
               <<"of dataset "<<eventset
               << Sas::error("MissingKeyword");
*/
      errstr<<"ccd number: "<<oalServer.ccd()
        <<Sas::message(Msg::App, Msg::Verbose);

  } else {
      errstr<<" Could not find dataset: "<< eventset
             << Sas::error("NoEventSet");
  }

 // Close the input dataset
  dataSetServer->close(evset);

}

/* Method: Epxrlcorr::calculateXRL()
*  Purpose: calculate the X-ray loading
*/
SImage Epxrlcorr::calculateXRL(string & offsetmap) 
{

  /* Open the offset map */
  DataSet * omset=0;

  if (dataSetServer->exists(offsetmap)) {
     omset = dataSetServer->open(offsetmap, DataSet::Read);
  } else {
     errstr<<" Error couldn't find offset map: "<< offsetmap
           << Sas::error("NoOffsetMap");
  }

  /* Read in the offset data from extension 1 */
  
  if (!omset->hasBlock("PNODI1")) {
     errstr<<" Error couldn't find PNODI1 "<< offsetmap
           << Sas::error("NoOffsetData");
  }

  /* Read in the offset data header */
  Array * array = omset->array(offsetmap+":PNODI1");
  ArrayData *ad = array->data();
  vector<unsigned long> extents = array->size();

  // Handle case where array is 3-d - just use first element of 3rd dimension
  // Dimensions are defined such that ncol is the final dimension and 
  // nrow the penultimate
  unsigned nrow = extents[extents.size()-2], ncol=extents[extents.size()-1];

  errstr<<"Offset map extents: "<<nrow<<" "<<ncol<<" dim: "<<extents.size()
          <<Sas::message(Msg::App, Msg::Verbose);

  SImage offset_map(nrow, ncol, ad->int16Data());
  SImage offset_map_master = offset_map;

  errstr<<"Set up simage [13][0]: "<<offset_map[13][0]
        <<Sas::message(Msg::App, Msg::Verbose);
  /*
   * Calculate the X-ray loading per pixel 
  */

  // Subtract the column medians from the non-zero elements
  vector<short int>tempVec;
 // unsigned nrow = extents[0], ncol=extents[1];

  for (unsigned c=0; c<ncol; ++c) {

  // Set up column vector
    for (unsigned r=0; r<nrow; ++r) {
       
      // make vector per column from non-zero values
       if (offset_map_master[r][c] != 0)
           tempVec.push_back(offset_map[r][c]);
    } 

    // Calc median value for this column
    short int median = getmedian(tempVec); 
   
    errstr<<"Median for col: "<<c<<" = "<<median
          <<Sas::message(Msg::App, Msg::Verbose);
    
    // Subtract median from each non-zero element of this column
    for (unsigned r=0; r<nrow; ++r) {
       if (offset_map_master[r][c] != 0)
          offset_map[r][c] -= median;
    }
    // Empty vector
    tempVec.erase(tempVec.begin(),tempVec.end());
  }

  // Subtract row medians
  for (unsigned r=0; r<nrow; ++r) {
    for (unsigned c=0; c<ncol; ++c) {
   
      // make vector per row where element !=0
       if (offset_map_master[r][c] != 0)
           tempVec.push_back(offset_map[r][c]);
    }

    // Calc median value for this row
    short int median = getmedian(tempVec);
   
    errstr<<"Median for row: "<<r<<" = "<<median
          <<Sas::message(Msg::App, Msg::Verbose);
    

    // Subtract median from each element of this row
    for (unsigned c=0; c<ncol; ++c) {
       if (offset_map_master[r][c] != 0)
          offset_map[r][c] -= median;
    }
    // Empty vector
    tempVec.erase(tempVec.begin(),tempVec.end());
  }

 // Close the offset map file
  dataSetServer->close(omset);

  return offset_map;
}

/* Method: getmedian()
*  Purpose: finds the median 1-d array
*/
short int Epxrlcorr::getmedian(vector<short int> tempVec) 
{
   // Return 0 if zero size vector
    if (tempVec.size() == 0)
       return 0;
   
   // Trick to find median 
   /*
    vector<short int>::iterator first = tempVec.begin();
    vector<short int>::iterator last  = tempVec.end();
    vector<short int>::iterator middle = first + (last - first) / 2;

    short int mid = *middle;
   */
    std::sort (tempVec.begin(), tempVec.end());

    return tempVec[floor(tempVec.size()/2)];
}

/* Method: XRLCorrApply()
*  Purpose: corrects event energies for X-ray loading
*/
void Epxrlcorr::XRLCorrApply(DataSet * evset, SImage XRLCorrMap)
{


   errstr<<"Applying correction [13][0]: "<<XRLCorrMap[13][0]
        <<Sas::message(Msg::App, Msg::Verbose);

  // Open the EVENTS block of the raw events file
   Table * evtab = 0;
   Column * enCol = 0;
   int offset=0; // Raw coords go from 0-63 (raw) or 1-64 (epframes(

   if (_rawinputfile) { // Raw event file inupt
      evtab = evset->table("PNIME1");
      enCol = evtab->column("ENERGY");
      offset=0;
   } else { //epframes input
      evtab = evset->table("EVENTS");
      enCol = evtab->column("PHA");
      offset=1;
   }

  // Get the energy and position arrays
   vector<unsigned> rawx, rawy;

   CalOalUtility::fillVector(evtab, "RAWX", rawx);
   CalOalUtility::fillVector(evtab, "RAWY", rawy);
  
   ColumnData * enColData = enCol->data();
   int16      * energy = enColData->int16Data();

  // Make a new column to contain the original energies
   Column * eorigCol = evtab->addColumn("ENERGY_ORIG", Column::Int16,
                                          "Original energy","CHANNEL");
   ColumnData * eorigColeData = eorigCol->data();
   int16      * eorig = eorigColeData->int16Data();

  // Loop over each charge 
   for (unsigned i=0; i<rawx.size(); ++i) {

      // Copy into the original energy column
       eorig[i] = energy[i];

      // Correct the event energy for X-ray loading
      // NB: If applied on the event files output by epframes then
      // need to subtract one from the RAWX/Y values of the events here.
       energy[i] += XRLCorrMap[rawy[i]-offset][rawx[i]-offset];

      //DBG
       //if (rawx[i] == 0) 
         // cout<<"R0: "<<(rawx[i])<<" "<<(rawy[i])<<" "<<XRLCorrMap[rawy[i]][rawx[i]]<<endl;
   }

   errstr<<"Energy of event one changed from: "<<eorig[0]
         <<" to "<<energy[0]<<Sas::message(Msg::App, Msg::Verbose);

  // Add a keyword stating that the correction has been applied
   *evtab->addAttribute("XRL_CORR", 
                 "Has X-ray loading correction been applied")="YES";

  // Close down
   delete eorigColeData; delete enColData;
}

/* Method: findOffsetMap()
*  Purpose: Find the offset map applicable to this event file
*/
string Epxrlcorr::findOffsetMap()
{

  // Get the offset map file name from input parameter if set 
   string offsetmap;
   if (booleanParameter("withoffsetmap"))

      offsetmap = stringParameter("offsetmap");

   else { // find the offset map file name from the ODF and CCD number

      const Odf *odf = oalServer.odf();
      if (odf->hasSet(Xmm::OFFSET_DATA))
         offsetmap = odf->selectSet(Xmm::OFFSET_DATA)->name();
      else
         errstr<<"OFFSET map data not found for event file "
               << Sas::error("NoOffsetMap");
   }

   errstr<<"Using offset map: "<<offsetmap
        <<Sas::message(Msg::App, Msg::Sparse);

   return offsetmap;
}

